from sklearn.base import TransformerMixin
import numpy as np
import pandas as pd
import optuna

class model_select(TransformerMixin):
    def __init__(self,transform_pipeline,X_test,y_test,n_trial=20,optuna_param=None,optuna_direction='minimize',classifier=False,estimator=False):
        """
        Parameters
        ----------
        transform_pipeline :sklearn.pipeline.Pipeline TYPE
            Pipeline module without model steps.
        X_test : Series TYPE
            X_test data.
        y_test : Series TYPE
            y_test data.
        n_trial : Int TYPE, optional
            n_trial variable for Optuna. The default is 20.
        optuna_param : dict, optional
            Hyperparamaters dict for optuna.If you don't give value,the parameter search function will run first. The default is None.
        optuna_direction : String, optional
            Minimize or Maximize for Optuna model Score. The default is 'minimize'.
        classifier : String, optional
            The model problem you want to establish is Regression or Classifier. The default is False.
        estimator : Object, optional
            If you specify an external estimator, LazyEstimator will not work. The default is False.
        """
        
        self.classifier=classifier
        self.pipe=transform_pipeline
        self.X_test=X_test
        self.y_test=y_test
        self.optuna_param=optuna_param
        self.optuna_direction=optuna_direction
        self.estimator=estimator
        self.n_trial=n_trial
    def prt(self,title=''):
        """
        This function using to console printing.

        Parameters
        ----------
        title : String TYPE, optional
             The default is ''.
        """
        print("=" * 2*len(title))
        c_title=title.center(2*len(title))
        print(c_title)
        print("=" * 2*len(title))
    def pipe_fit(self,X,y):
        """
        For the defined pipeline, it transforms X_test appropriately through the fit pipeline and prepares it for optuna.    

        Parameters
        ----------
        X : Series,DataFrame TYPE
            X_train.
        y : Series,DataFrame
            y_train (target).

        Returns
        -------
        X_train_trans : Series TYPE
            Transformed X_train.
        X_test_trans :Series TYPE
            Transformed X_test.

        """
        from sklearn import set_config            
        self.prt('Pipe Transformation Started... ')
        set_config(display="diagram")
        pipe_transform=self.pipe
        X_train_trans=pipe_transform.fit_transform(X,y)
        X_test_trans=pipe_transform.transform(self.X_test)
        return X_train_trans,X_test_trans
    class find_params:
        def __init__(self,est):
            """
            It is intended for finding parameters for an estimator found or defined by Lazyestimator and for which no parameter is defined.
            Parameters
            ----------
            est : Object TYPE
                Estimator.

            """
            self.est=est
            self.ridge_params={}
            self.lasso_params={}
            self.elasticnet_params={}
            self.dt_params={}
            self.randomforest_params={}
            self.gradient_params={}
            self.xgb_params={}
            self.lgbm_params={}
            self.catboost_params={}
            self.reg_result=self.reg()
            self.clf_result=self.clf()
        def reg(self):
            """
            This Function define to parameters for Regression Estimators.
            
            """
            
            "Regressor Models"
            from sklearn.linear_model import Ridge 
            from sklearn.linear_model import Lasso 
            from sklearn.linear_model import ElasticNet 
            from sklearn.tree import DecisionTreeRegressor 
            from sklearn.ensemble import RandomForestRegressor 
            from sklearn.ensemble import GradientBoostingRegressor 
            from xgboost import XGBRegressor 
            from lightgbm import LGBMRegressor 
            from sklearn.svm import SVR 
            from catboost import CatBoostRegressor 
            if self.est == Ridge:
                self.ridge_params={'alpha': [0.1, 1.0, 10.0], 'solver': ['auto', 'svd', 'cholesky', 'lsqr']}
                return self.ridge_params
            elif self.est == Lasso:
                self.lasso_params={'alpha': [0.1, 1.0, 10.0], 'max_iter': [1000, 2000, 5000], 'selection': ['cyclic', 'random']}
                return self.lasso_params
            elif self.est == ElasticNet:
                self.elasticnet_params={'alpha': [0.1, 1.0, 10.0], 'l1_ratio': [0.25, 0.5, 0.75], 'max_iter': [1000, 2000, 5000]}
                return self.elasticnet_params
            elif self.est == DecisionTreeRegressor:
                self.dt_params={'max_depth': [None, 5, 10], 'min_samples_split': [2, 5, 10], 'min_samples_leaf': [1, 2, 3]}
                return self.dt_params
            elif self.est == RandomForestRegressor:
                self.randomforest_params={'n_estimators': [100, 200, 500], 'max_depth': [None, 5, 10], 'min_samples_split': [2, 5, 10]}
                return self.randomforest_params
            elif self.est == GradientBoostingRegressor:
                self.gradient_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.gradient_params
            elif self.est == XGBRegressor:
                self.xgb_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.xgb_params
            elif self.est == LGBMRegressor:
                self.lgbm_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.lgbm_params
            elif self.est == CatBoostRegressor:
                self.catboost_params={'iterations': [100, 200, 500],'learning_rate': [0.01, 0.1, 0.2],'depth': [4, 6, 8],'l2_leaf_reg': [1, 3, 5],
                        'border_count': [32, 64, 128]}
                return self.catboost_params
            elif self.est == SVR:
                self.svr_params={'C': [0.1, 1.0, 10.0], 'epsilon': [0.1, 0.01, 0.001], 'kernel': ['linear', 'rbf'], 'gamma': ['scale', 'auto']}
                return self.svr_params
            return []
        def clf(self):
            """
            This Function define to parameters for Regression Estimators.
   
            """
            "Classification Models"
            from sklearn.linear_model import LogisticRegression 
            from sklearn.neighbors import KNeighborsClassifier 
            from sklearn.svm import SVC 
            from sklearn.tree import DecisionTreeClassifier 
            from sklearn.ensemble import RandomForestClassifier 
            from sklearn.ensemble import GradientBoostingClassifier 
            from xgboost import XGBClassifier 
            from lightgbm import LGBMClassifier 
            from catboost import CatBoostClassifier 
            from sklearn.naive_bayes import GaussianNB
            if self.est == LogisticRegression:
                self.logreg_params={'C': [0.1, 1.0, 10.0], 'solver': ['liblinear', 'lbfgs'], 'max_iter': [100, 200, 500]}
                return self.logreg_params
            elif self.est == KNeighborsClassifier:
                self.knn_params={'n_neighbors': [3, 5, 10], 'weights': ['uniform', 'distance'], 'p': [1, 2, 3]}
                return self.knn_params
            elif self.est == SVC:
                self.svc_params={'C': [0.1, 1.0, 10.0], 'kernel': ['linear', 'rbf'], 'gamma': ['scale', 'auto'], 'decision_function_shape': ['ovo', 'ovr']}
                return self.svc_params
            elif self.est == DecisionTreeClassifier:
                self.dt_params={'max_depth': [None, 5, 10], 'min_samples_split': [2, 5, 10], 'criterion': ['gini', 'entropy']}
                return self.dt_params
            elif self.est == RandomForestClassifier:
                self.randomforest_params={'n_estimators': [100, 200, 500], 'max_depth': [None, 5, 10], 'min_samples_split': [2, 5, 10], 'criterion': ['gini', 'entropy']}
                return self.randomforest_params
            elif self.est == GradientBoostingClassifier:
                self.gradient_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.gradient_params
            elif self.est == XGBClassifier:
                self.xgb_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.xgb_params
            elif self.est == LGBMClassifier:
                self.lgbm_params={'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1, 0.2], 'max_depth': [3, 5, 7]}
                return self.lgbm_params
            elif self.est == CatBoostClassifier:
                self.catboost_params={'iterations': [100, 200, 500],'learning_rate': [0.01, 0.1, 0.2],'depth': [4, 6, 8],'l2_leaf_reg': [1, 3, 5],
                        'border_count': [32, 64, 128]}
                return self.catboost_params
            elif self.est == GaussianNB:
                self.svr_params={}
                return self.svr_params
            return []
        def convert_params(self,param_dict):
            """
            Converts the found parameters into a dictionary of parameters suitable for the Optuna (Trial) object.

            Parameters
            ----------
            param_dict : Dict TYPE
                Paramter dict basicly not have a trial object.

            Returns
            -------
            Dict TYPE
                Returns a new dict according to the Trial object.

            """
            def o_params(trial):
                params_c={}
                for key,value in param_dict.items():
                    params_c[key]=trial.suggest_categorical(key,value)
                return params_c
            return o_params

    def lazy_model(self,X,y):
        """
        Returns the estimator that gives the best score among the estimators defined in the list.


        Parameters
        ----------
        X : Series / DataFrame TYPE
            DESCRIPTION.
        y : Series / DataFrame TYPE
            DESCRIPTION.

        Returns
        -------
        Object TYPE
            Estimator Returned.

        """
        
        "Regressor Models"
        from sklearn.linear_model import Ridge 
        from sklearn.linear_model import Lasso 
        from sklearn.linear_model import ElasticNet 
        from sklearn.tree import DecisionTreeRegressor 
        from sklearn.ensemble import RandomForestRegressor 
        from sklearn.ensemble import GradientBoostingRegressor 
        from xgboost import XGBRegressor 
        from lightgbm import LGBMRegressor 
        from sklearn.svm import SVR 
        from catboost import CatBoostRegressor 
        "Classification Models"
        from sklearn.linear_model import LogisticRegression 
        from sklearn.neighbors import KNeighborsClassifier 
        from sklearn.svm import SVC 
        from sklearn.tree import DecisionTreeClassifier 
        from sklearn.ensemble import RandomForestClassifier 
        from sklearn.ensemble import GradientBoostingClassifier 
        from xgboost import XGBClassifier 
        from lightgbm import LGBMClassifier 
        from catboost import CatBoostClassifier 
        from sklearn.naive_bayes import GaussianNB

        from sklearn.metrics import f1_score,mean_squared_error

        verbose_list=[GradientBoostingRegressor,
                            XGBRegressor,
                            LGBMRegressor,
                            CatBoostRegressor,
                            GradientBoostingClassifier,
                            XGBClassifier,
                            LGBMClassifier,
                            CatBoostClassifier]

        regressor=[Ridge,Lasso,ElasticNet,DecisionTreeRegressor,RandomForestRegressor,GradientBoostingRegressor,
                  XGBRegressor,LGBMRegressor,SVR,CatBoostRegressor]
        classifications=[LogisticRegression,KNeighborsClassifier,SVC,DecisionTreeClassifier,RandomForestClassifier,
                        GradientBoostingClassifier,XGBClassifier,LGBMClassifier,GaussianNB,CatBoostClassifier]
        results=[]
        if self.estimator==False:
            self.prt("Lazy Best Estimator Searching Started...")
            if self.classifier==True:
                for clf in classifications:
                    model=clf()
                    model.fit(self.X_train_trans,y)
                    y_pred=model.predict(self.X_test_trans)
                    f1 = f1_score(self.y_test, y_pred, average='weighted')
                    results.append(f1)
                results=np.array(results)
                max_index = np.argmax(results)
                self.prt('Best Estimator')
                print(classifications[max_index])
                self.prt('              ')
                return classifications[max_index]
            else:
                for est in regressor:
                    model=est()
                    model.fit(self.X_train_trans,y)
                    y_pred=model.predict(self.X_test_trans)
                    mse=mean_squared_error(self.y_test,y_pred)
                    results.append(mse)
                results=np.array(results)
                min_index = np.argmin(results)
                self.prt('Best Estimator')
                print(regressor[min_index])
                return regressor[min_index],verbose_list
    
    def fit(self,X,y):
        from functools import partial
        from sklearn.model_selection import cross_val_score
        from sklearn.metrics import mean_squared_error,classification_report,f1_score

        def objective(trial,best_est,X_train,y_train,X_test,y_test,param,verbose):
            def convert_object_param(trial,model):
                """
                It translates the parameter values to " max_depth:trial.suggest_categorical" and maps the parameters to the model.

                Parameters
                ----------
                trial : TYPE
                    DESCRIPTION.
                model : TYPE
                    DESCRIPTION.

                Returns
                -------
                TYPE
                    DESCRIPTION.

                """
                
                params={}
                for key,value in param(trial).items():
                    params[key]=value
                return model(**params)

            model=convert_object_param(trial,best_est) # Returns the model with parameters mapped together with the best estimator.
            if (model=='xgboost.sklearn.XGBRegressor') | (model=='xgboost.sklearn.XGBClassifier'):
                eval_set = [(X_train, y_train), (X_test, y_test)]
                model.fit(X_train,y_train,eval_set=eval_set,early_stopping_rounds=10, verbose=0)
            else:
                model.fit(X_train,y_train)
            y_pred=model.predict(X_test)
            #avg_score=mean_squared_error(y_pred,y_test)
            # Classifier and regression model metrics are reserved for metrics.
            if self.classifier==True: # 
                f1 = f1_score(y_test, y_pred, average='weighted')
                report = classification_report(y_test, y_pred)
                print(report)
                return f1
            else:
                scores = cross_val_score(model, X_train, y_train, cv=5,scoring='neg_mean_squared_error')
                avg_score = scores.mean()
                return -avg_score
        
        self.X_train_trans,self.X_test_trans=self.pipe_fit(X,y) # Pipeline fit and transform Starts
        best_est,v_list=self.lazy_model(self.X_train_trans,y) # Lazyestimator starts
        #If parameters not defined , call the before defined finds parameters
        if self.optuna_param==None: 
            if self.find_params(best_est).reg_result !=[]:
                param=self.find_params(best_est).convert_params(self.find_params(best_est).reg_result)
            else:
                param=self.find_params(best_est).convert_params(self.find_params(best_est).clf_result)

        else:
            param=self.optuna_param
            
        self.prt('Optuna Has Started...')
        study = optuna.create_study(direction=self.optuna_direction) 
        objective_with_param=partial(objective,best_est=best_est,X_train=self.X_train_trans,y_train=y,X_test=self.X_test_trans,
                                     y_test=self.y_test,param=param,verbose=v_list)

        
        study.optimize(objective_with_param, n_trials=self.n_trial) # Optuna start
        best_params = study.best_params
        self.prt('Best Params')
        print(best_params) #Print best params
        self.model=best_est(**best_params)
        self.model.fit(self.X_train_trans,y)
        
        return self
    def predict(self):
        """
        Predicts X_test_trans previously transformed from the pipeline and returns predict.

        Returns
        -------
        pred :Array TYPE

        """
        pred=self.model.predict(self.X_test_trans)
        return pred


from lazyauto.Eda.eda import dedect_features,skewness,null_val
from sklearn import datasets
import pandas as pd

diab=datasets.load_diabetes(as_frame=True)
df=pd.DataFrame(diab['frame'],columns=diab['feature_names']+['target'])

def test_dedect_features():
    date_cols,num_but_cat,cat_col,cat_but_num,num_col=dedect_features(df)
    assert (date_cols,num_but_cat,cat_col,cat_but_num,num_col) == ([],['sex'],[],[],['age', 'bmi', 'bp', 's1', 's2', 's3', 's4', 's5', 's6', 'target'])
def test_skewness():
    assert skewness(df) == ['bmi','s3','s4']
def test_null_val():
    assert any(null_val(df)[0]==0)


import pandas as pd
from lazyauto.time_transform.date_time_transform import datetime_ohe, datetime_simple

start_date = pd.to_datetime("2023-03-08")
end_date = pd.to_datetime("2023-03-15")

dates = pd.date_range(start_date, end_date)
df = pd.DataFrame(dates, columns=["Date"])

def test_datetime_ohe():
    a= datetime_ohe(df['Date'])
    pd.isnull(a).sum()
    assert any(pd.isnull(a).sum() == 0)
    
def test_datetime_simple():
    a= datetime_simple(df['Date'])
    pd.isnull(a).sum()
    assert any(pd.isnull(a).sum() == 0)
    


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.figure_factory as ff
import plotly.express as px
import random
from time import sleep
import plotly.io as io
io.templates.default = 'plotly_dark'



def show(df, caption=""):
    """Helper function to set dataframe presentation style Html Based like JupiterNotebook.""" # noqa:E501
    return df.style.background_gradient(cmap='Reds').set_caption(caption).set_table_styles([{
        'selector': 'caption',
        'props': [
            ('color', 'Red'),
            ('font-size', '18px'),
            ('font-weight', 'bold')
        ]}])


def skewness(df):
    """
    This function find skewness above 0.5 and return list. # noqa:E501

    Parameters
    ----------
    df : DataFrame

    Returns
    -------
    TYPE
        list.

    """
    skewness=df.skew()
    return skewness[skewness>0.5].index.to_list()


def null_val(df):
    """
    This function find if have null value. # noqa:E501

    Parameters
    ----------
    df : DataFrame

    Returns
    -------
    TYPE 
        Dataframe
    """
    return df.isna().sum().to_frame()


def describe_object(df):
    """
    This function call describe and just other than "number" and return. # noqa:E501

    Parameters
    ----------
    df : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.
    """
    return df.describe(exclude='number').T
def describes(df):
    return df.describe(percentiles=[0.01,0.05, 0.1 , 0.25,  0.5,  0.75, 0.9 , 0.99]).T


def unique_val(df_train,df_test,cat_col):
    """
    

    Parameters
    ----------
    df_train : DataFrame
        Train Data.
    df_test : DataFrame
        Test Data.
    cat_col : list
        Catagoric columns in data.

    Returns
    -------
    v_df : TYPE
        DESCRIPTION.
    compare : TYPE
        DESCRIPTION.

    """
    compare={}
    v_df=pd.DataFrame()
    for ind,col in enumerate(cat_col):  
        a=list(df_train[col].unique())
        b=list(df_test[col].unique())
        res = [x for x in b if x not in a]
        compare[col]=res
        v_df.loc[0,col]=len(res)
    return v_df,compare


def dedect_features(df,show=False,nc_th=20,cn_th=30):
    '''
    This function dedect features type and return as a list. # noqa:E501
    
        Parameters
        ----------
        df : DataFrame
            Main Df
        nc_th : Integer, 
            Numeric but categorical threshold limit. The default is 20.
        cn_th : Integer, 
            Categorical but numeric threshold limit. The default is 30.

        Returns
        -------
        date_cols : List
            Date Types Columns List.
        num_but_cat : List
            Numeric but Categorical Types Columns List.
        cat_col : List
            Categorical Types Columns List.
        cat_but_num : List
            Categorical but Numeric Columns List.
        num_col : List
            Date Types Columns List.

    '''
    date_cols=[col for col in df if df[col].dtypes==("datetime64[ns]")]

    num_but_cat=[col for col in df.select_dtypes(['float','integer']).columns if df[col].value_counts().shape[0]<nc_th]
    cat_col=df.select_dtypes(['object','category']).columns.to_list()
    cat_but_num=[col for col in df.select_dtypes(['object','category']).columns if df[col].value_counts().shape[0]>cn_th]
    cat_col=[col for col in cat_col if col not in cat_but_num]

    num_col=df.select_dtypes(['integer','float']).columns.to_list()
    num_col=[col for col in num_col if col not in num_but_cat]
    if show==True:
        print("Date Columns: ",date_cols)
        print("Numeric But Categoric Columns: ",num_but_cat)
        print("Categoric Columns: ",cat_col)
        print("Categoric But Numeric Columns: ",cat_but_num)
        print("Numeric Columns: ",num_col)
    
    return date_cols,num_but_cat,cat_col,cat_but_num,num_col
    

def corr(df,title):
    """
    Parameters
    ----------
    df : DataFrame
        
    title : String
        Heatmap Title.
    """
    
    corrr=df.astype(float).corr()
    colormap = plt.cm.RdBu_r
    plt.figure(figsize=(15, 15))
    plt.title(f'{title} Correlation of Features', fontweight='bold', y=1.02, size=20)
    sns.heatmap(corrr, linewidths=0.1, vmax=1.0, vmin=-1.0,mask=np.triu(np.ones_like(corrr),k=1), 
        square=True, cmap=colormap, linecolor='white', annot=True, annot_kws={"size": 11, "weight": "bold"},fmt=".2f",
               cbar_kws={"shrink":0.8})


def cat_plot(df,target,title='',drop=[]):
    """
    Draws the distrubiton graph of categoric values ​​with piechart and bar plot.# noqa:E501


    Parameters
    ----------
    df : DataFrame
        
    target : String
        Target Features in data.
    title : String, optional
        Subplot Title. The default is ''.
    drop : List, optional
        If you want drop any feature , you have to defined list. The default is [].# noqa:E501
        
    """
    date_cols,num_but_cat,cat_col,cat_but_num,num_col=dedect_features(df,20,30)
    cols=df[cat_col+num_but_cat].columns
    n_cols=2
    n_rows=len(cols)

    fig, axes = plt.subplots(n_rows, n_cols, figsize=(20, 5*len(cols)),squeeze=False)
    plt.rcParams['figure.facecolor']='#232425'
    plt.rcParams['text.color'] = 'w'
    plt.rcParams['xtick.color'] = 'w'
    plt.rcParams['ytick.color'] = 'w'
    plt.rcParams['axes.labelcolor'] = 'w'
    plt.rcParams['axes.edgecolor'] = 'w'
    plt.rcParams['axes.titlesize']=20.0
    plt.rcParams['axes.labelsize']=15.0
    plt.rcParams['axes.titlepad']=20.0
    plt.rcParams['axes.titlecolor']='r'
    plt.rcParams['axes.titleweight']=15.0
    plt.rcParams['font.size']=11.0
    

    for i, var_name in enumerate(cols,start=1):
        i-=1
        grouped=df.groupby([var_name],as_index=False)[target].sum()
        _, wedges, autotexts = axes[i,0].pie(grouped[target], labels=grouped[var_name], autopct='%1.1f%%', startangle=10, colors=sns.color_palette('pastel'),
                                        textprops={'fontsize': 14,'va':"center",'rotation_mode' : 'anchor'},rotatelabels=True,wedgeprops={"linewidth":1.5,"edgecolor":"white"})
        for autotext in autotexts:
            autotext.set_color('black')
            autotext.set_fontsize(15)
        axes[i,0].set_title(var_name)
        axes[i, 0].set_facecolor('#232425')
        axes[i, 1].set_facecolor('#232425')
        ax=sns.barplot(x=var_name, y=target, data=grouped, ax=axes[i, 1],palette=sns.color_palette('pastel'))
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right',size=15)
        ax.set_yticklabels(ax.get_yticklabels(),size=15)
        ax.set_xlabel('')
        ax.spines['left'].set_color('white')        
        ax.spines['top'].set_color('white')
        ax.spines['right'].set_color('white')
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_lw(1)
        ax.spines['top'].set_lw(1)
        ax.spines['right'].set_lw(1)
        ax.spines['bottom'].set_lw(1)
        axes[i,1].set_title(var_name)

    plt.subplots_adjust(bottom=0.3, wspace=0.2)
    plt.show()
   
    
def pairplot(df,title,n_cols=3,drop=[],render_local='False'):
    """
    Draws the distrubiton graph of continuous values ​​with box plot.# noqa:E501


    Parameters
    ----------
    df : DataFrame
    title : TYPE
        Figure Title.
    n_cols : Int, optional
        Subplots column number. The default is 3.
    drop : TYPE, optional
       If you want drop any feature , you have to defined list. The default is [].# noqa:E501

    """
    if render_local=='True': io.renderers.default = 'browser'
    date_cols,num_but_cat,cat_col,cat_but_num,num_col=dedect_features(df,20,30)
    colors_d = '#E48F72'
    cols=df[num_col].columns.drop(drop).to_list()
    n_rows=((len(cols) - 1) // n_cols + 1)*2
    
    cnt=0
    title_c=0
    sub_title=[]
    for i in range(0,len(cols)*2-1):
        if (cnt>2) & (cnt<6):
            sub_title.append("")
            cnt+=1
            if cnt==6: 
                cnt=0
            continue
        else:
            if len(cols)-1>=title_c:
                sub_title.append("<span style='color:#d61c38;font-weight:bold;'>{}</span>".format(cols[title_c]))
                title_c+=1
                cnt+=1
            else:
                break
    fig= make_subplots(rows=n_rows,cols=n_cols,subplot_titles=(sub_title))
    col_n=0
    row_n=1
    colors_list = px.colors.qualitative.Alphabet
    selected_elements=[]
    for i, var_name in enumerate(cols,start=1):
        def random_element_without_duplicates(list):
            """
            Return a random element from the list without duplicates.# noqa:E501
          
            Args:
              list: The list of elements.
          
            Returns:
              A random element from the list.
            """
            
            random_index = random.randint(0, len(list) - 1)
            while list[random_index] in selected_elements:
                random_index = random.randint(0, len(list) - 1)
            selected_elements.append(list[random_index])
            
            return list[random_index]
        
        random_color = random_element_without_duplicates(colors_list)
        
        sleep(0.25)
        col_n=col_n+1
        if col_n==n_cols+1:
            col_n=1
            row_n=row_n+2
        if row_n==n_rows+1:
            row_n=1
            
        boxfig= go.Figure(data=[go.Box(x=df[var_name], showlegend=False, notched=True, marker_color=random_color, name=var_name)])
        for k in range(len(boxfig.data)):
            fig.add_trace(boxfig.data[k], row=row_n, col=col_n)
        group_labels = [var_name]
        hist_data=[df[var_name].values]
        distplfig = ff.create_distplot(hist_data, group_labels, colors=[random_color],
                     bin_size=.2, show_rug=False)
        for k in range(len(distplfig.data)):
            fig.add_trace(distplfig.data[k],
            row=row_n+1, col=col_n)
    fig.update_layout(barmode='overlay',
                     title={'text':f'<b>{title} Distrubitions</b>',
                             'font_size':24,
                             'font_color': 'Red',
                             'y':0.98,
                             'x':0.5,
                             'xanchor': 'center',
                             'yanchor': 'top'}
                     ,showlegend=False,height=1400,width=1200)
    colors_list = px.colors.qualitative.Alphabet
    fig.show()
  
    
def outlier_dedection(df,target,model,drop='',contaminations=[0.0001,0.0005,0.001,0.005,0.01,0.07,0.1,0.15,0.25],graph=False):
    """
    Function using Isolation Forest method and find the outlier data.# noqa:E501

    Parameters
    ----------
    df : DataFrame
        
    target : String
        Target Feature in DataFrame
    drop : List
        If you want drop any feature , you have to defined list.
    model : Object Funciton
        Define estimator for isolation forest.
    contaminations : List, optional
        The value that must be defined for outlier precision in the isolation forest. The default is [0.0001,0.0005,0.001,0.005,0.01,0.07,0.1,0.15,0.25].# noqa:E501
    graph : Bool, optional
        If you want draw contamination value of features,change statment to 'True' "". The default is False.# noqa:E501

    Returns
    -------
    list

    """
    
    
    
    n_cols=3
    from sklearn.ensemble import IsolationForest
    from lightgbm import LGBMRegressor
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import mean_squared_error
    from sklearn.decomposition import PCA
    cols=df.columns
    if drop!='':
        cols=df.columns.drop(drop)
    anomaly_dict={}
    n_rows=((len(cols) - 1) // n_cols + 1)
    col_n=0
    row_n=1
    fig= make_subplots(rows=n_rows,cols=n_cols,subplot_titles=(cols))
    def evaluate_outlier_classifier(model,data):
        labels=model.fit_predict(data)
        return data[labels==1],data[labels==-1]
    def evaluate_regressor(model,inliers,target):
        X=inliers.drop(target,axis=1)
        y=inliers[target]
        X_train, X_test, y_train, y_test = train_test_split(X, y,test_size=0.25,random_state=42)

        model.fit(X_train, y_train)
        preds = model.predict(X_test)
        rmse = mean_squared_error(y_test, preds, squared=False)
        return round(rmse, 3)
    
    scores = dict()
    for c in contaminations:
        # Instantiate IForest with the current c
        iforest = IsolationForest(n_estimators=100,max_samples='auto',max_features=1.0,bootstrap=False,contamination=c, random_state=42,verbose=0)

        # Get inliers with the current IForest
        inliers,outliers = evaluate_outlier_classifier(iforest, df[cols])

        # Calculate and store RMSE into scores
        rmse = evaluate_regressor(model,inliers,target)
        scores[c]=rmse

    print(scores)
    best_value = max(scores, key=scores.get)
    inliers,outliers = evaluate_outlier_classifier(iforest, df[cols])
    
    
    
    if graph==True:
        pca = PCA(2)
        pca.fit(df[cols])
        res=pd.DataFrame(pca.transform(df[cols]))
        plt.title("IsolationForest with PCA 2-Dimensions")
        
        b1 = plt.scatter(res[0], res[1], c='green',
                         s=20,label="normal points")
        b1 =plt.scatter(res.iloc[outliers.index.to_list(),0],res.iloc[outliers.index.to_list(),1], c='green',s=20,  edgecolor="red",label="predicted outliers")
        plt.legend(loc="upper right")
        plt.show()
        
        for col in cols:
            def setColor(val):
                if ([True for i in outliers.index.to_list() if i==val]):

                    return "red"
                else:
                    return "green"
            col_n=col_n+1
            if col_n==n_cols+1:
                col_n=1
                row_n=row_n+1
            if row_n==n_rows+1:
                row_n=1
            fig.add_trace(go.Scatter(
                            y=df[col],
                            mode='markers',
                            opacity=0.7,
                            marker=dict(color=[setColor(val) for val in df[col].index])
                        ),
                        row=row_n,col=col_n    
                              )
        fig.update_layout(barmode='overlay',
                        title={'text':'<b>Train_df Outliers with Isolation Forest</b>',
                                 'font_size':20,
                                 'font_color': 'Red',
                                 'y':0.98,
                                 'x':0.5,
                                 'xanchor': 'center',
                                 'yanchor': 'top'},showlegend=False,height=1200,width=1000)
        fig.show()
        
        
        
        
        
    return outliers




def drop_outliers(df,drop_index):
    """
    It drops the index list resulting from the outlier analysis from the dataframe.   

    Parameters
    ----------
    df :DataFrame TYPE
        
    drop_index : Dataframe TYPE
        DESCRIPTION.

    Returns
    -------
    df_out : TYPE
        DESCRIPTION.

    """
    f=df.shape[0]
    df_out=df.copy()
    print("Shape: ",f)
    for i in drop_index.index.to_list():
        try:
            df_out.drop(i,axis=0,inplace=True)
        except KeyError:
            continue
    df_out=df_out.reset_index(drop=True)
    l=df_out.shape[0]
    print("Later Dropped Outliers Shape: ",l)
    print("Result: ",f-l)
    print('Completed..')
    return df_out


import numpy as np
import pandas as pd
from lazyauto.model_select.lazy_model_ev import model_select
from lazyauto.Eda import eda
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.compose import ColumnTransformer

diab=datasets.load_diabetes(as_frame=True)
df=pd.DataFrame(diab['frame'],columns=diab['feature_names']+['target'])


date_cols,num_but_cat,cat_col,cat_but_num,num_col=eda.dedect_features(df,show=True)
ord_col=date_cols+num_but_cat+cat_but_num
num_col.remove('target')
X=df.drop(['target'],axis=1)
y=df['target']



X_train,X_test,y_train,y_test=train_test_split(X,y, \
    test_size=0.2, random_state=0)

    
    
from sklearn.preprocessing import OrdinalEncoder,StandardScaler,OneHotEncoder,FunctionTransformer
from sklearn.pipeline import Pipeline

cardinal_transformer=Pipeline(steps=[
    ('N_encoder',OrdinalEncoder(handle_unknown='use_encoded_value', unknown_value=-1))
])
numeric_transformer=Pipeline(steps=[
    ('scaler_robust',StandardScaler())
])
categoric_transformer=Pipeline(steps=[
    ('OHE',OneHotEncoder(drop='first',handle_unknown='ignore'))
])
func_trans=Pipeline(steps=[
    ('FT',FunctionTransformer(lambda x : np.where(x < 0, 0, np.log1p(x))))
])
preprocessing_transform=ColumnTransformer(transformers=[
    ('Numeric_trans',numeric_transformer,num_col),
    ('card_trans',cardinal_transformer,ord_col),
])
pipe=Pipeline([
    ('column_transformers',preprocessing_transform)
])

def optna_params(trial):
    params={
    'alpha' : trial.suggest_categorical('alpha', [1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 0.0, 1.0, 10.0, 100.0]),
    'l1_ratio' : trial.suggest_discrete_uniform('l1_ratio', 0.1, 1.0, 0.01)
    }
    return params
def test_lazy_model_ev():
    ml=model_select(transform_pipeline=pipe,X_test=X_test,y_test=y_test,n_trial=5,optuna_param=optna_params,optuna_direction='minimize',classifier=False)
    ml.fit(X_train,y_train)
    pred=ml.predict()
    assert pred != []


import pandas as pd
import numpy as np
def datetime_simple(df_date):
    """
    Datetime column transform and seperate to day,month,year,week,dayofweek ... etc.

    Parameters
    ----------
    df_date : DataFrame / Series TYPE
        Datetime DataFrame/Series.

    Returns
    -------
    DataFrame TYPE
        
    """
    def is_spring(df):
        date = pd.to_datetime(df)
        return (date.month >= 2) & (date.month <= 4)
    def is_summer(df):
        date = pd.to_datetime(df)
        return (date.month >= 5) & (date.month <= 7)
    def is_autumn(df):
        date = pd.to_datetime(df)
        return (date.month >= 8) & (date.month <= 10)
    def is_winter(df):
        date = pd.to_datetime(df)
        return (date.month >= 11) | (date.month <= 1)
    df_copy=pd.DataFrame(df_date.values,columns=['DATE'])
    df_copy['DATE']=pd.to_datetime(df_copy['DATE'])
    df_copy['day']=df_copy['DATE'].dt.day
    df_copy['month']=df_copy['DATE'].dt.month
    df_copy['year']=df_copy['DATE'].dt.year
    df_copy['week']=df_copy['DATE'].dt.week
    df_copy['dayofweek']=df_copy['DATE'].dt.dayofweek
    df_copy['weekend']=[1 if i>=5 else 0 for i in df_copy['DATE'].dt.weekday]
    df_copy['quarter']=df_copy['DATE'].dt.quarter
    df_copy['month_start']=df_copy['DATE'].dt.is_month_start.apply(lambda x : 1 if x==True else 0)
    df_copy['month_end']=df_copy['DATE'].dt.is_month_end.apply(lambda x : 1 if x==True else 0)
    df_copy['leap_year']=df_copy['DATE'].dt.is_leap_year.apply(lambda x : 1 if x==True else 0)
    df_copy['Q-MAR'] = df_copy['DATE'].dt.to_period('Q-MAR')
    df_copy['fiscal year'] = df_copy['Q-MAR'].dt.qyear
    df_copy['Q-MAR'] = df_copy['DATE'].dt.to_period('Q-MAR').astype("object")
    df_copy['ymd'] = (df_copy['year']*7)*10000+df_copy['month']*12*100+df_copy['day']*30
    df_copy['sin'] = np.sin(2 * np.pi * df_copy['ymd']/365)
    df_copy['cos'] = np.cos(2 * np.pi * df_copy['ymd']/365)
    df_copy['is_spring'] = df_copy['DATE'].apply(is_spring).astype(int)
    df_copy['is_summer'] = df_copy['DATE'].apply(is_summer).astype(int)
    df_copy['is_autumn'] = df_copy['DATE'].apply(is_autumn).astype(int)
    df_copy['is_winter'] = df_copy['DATE'].apply(is_winter).astype(int)

    return df_copy
def datetime_ohe(df_date):
    """
    All these transformed values transform to like OneHotEncoder type all features. All unique values is new features in the DataFrame. 

    Parameters
    ----------
    df_date : DataFrame / Series TYPE
        Datetime DataFrame/Series.

    Returns
    -------
    TYPE

    """
    def is_spring(df):
        date = pd.to_datetime(df)
        return (date.month >= 2) & (date.month <= 4)
    def is_summer(df):
        date = pd.to_datetime(df)
        return (date.month >= 5) & (date.month <= 7)
    def is_autumn(df):
        date = pd.to_datetime(df)
        return (date.month >= 8) & (date.month <= 10)
    def is_winter(df):
        date = pd.to_datetime(df)
        return (date.month >= 11) | (date.month <= 1)
    clmns=['day','month', 'dayofweek', 'hour']
    other_cols = [f'day_{i}' for i in range(1,32)]+[f'dayofweek{i}' for i in range(1, 8)] + [f'hour_{i}' for i in range(1, 25)] + [f'month_{i}' for i in range(1, 13)]

    df_copy=pd.DataFrame(df_date.values,columns=['DATE'])
    df_copy['DATE']=pd.to_datetime(df_copy['DATE'])
    df_copy['day']=df_copy['DATE'].dt.day
    df_copy['month']=df_copy['DATE'].dt.month
    df_copy['year']=df_copy['DATE'].dt.year
    df_copy['week']=df_copy['DATE'].dt.week
    df_copy['dayofweek']=df_copy['DATE'].dt.dayofweek
    df_copy['weekend']=[1 if i>=5 else 0 for i in df_copy['DATE'].dt.weekday]
    df_copy['quarter']=df_copy['DATE'].dt.quarter
    df_copy['month_start']=df_copy['DATE'].dt.is_month_start.apply(lambda x : 1 if x==True else 0)
    df_copy['month_end']=df_copy['DATE'].dt.is_month_end.apply(lambda x : 1 if x==True else 0)
    df_copy['leap_year']=df_copy['DATE'].dt.is_leap_year.apply(lambda x : 1 if x==True else 0)
    df_copy['Q-MAR'] = df_copy['DATE'].dt.to_period('Q-MAR')
    df_copy['fiscal year'] = df_copy['Q-MAR'].dt.qyear
    df_copy['Q-MAR'] = df_copy['DATE'].dt.to_period('Q-MAR').astype("object")
    df_copy['ymd'] = (df_copy['year']*7)*10000+df_copy['month']*12*100+df_copy['day']*30
    df_copy['sin'] = np.sin(2 * np.pi * df_copy['ymd']/365)
    df_copy['cos'] = np.cos(2 * np.pi * df_copy['ymd']/365)
    df_copy['is_spring'] = df_copy['DATE'].apply(is_spring).astype(int)
    df_copy['is_summer'] = df_copy['DATE'].apply(is_summer).astype(int)
    df_copy['is_autumn'] = df_copy['DATE'].apply(is_autumn).astype(int)
    df_copy['is_winter'] = df_copy['DATE'].apply(is_winter).astype(int)

    hour=df_copy['DATE'].dt.hour.value_counts()
    if hour[0]!=df_copy.shape[0]:
        df_copy['hour']=df_copy['DATE'].dt.hour
    else:
        other_cols = [f'day_{i}' for i in range(1,32)]+[f'dayofweek{i}' for i in range(1, 8)] + [f'month_{i}' for i in range(1, 13)]
        clmns.remove('hour')
    for column in clmns:
        index = [col for col in df_copy.columns if col!=column]
        df_copy = df_copy.pivot_table(index=index, columns=[column], aggfunc=np.count_nonzero).fillna(0).astype(bool).add_prefix(f'{column}_').reset_index()

    df_base = pd.DataFrame(columns= ['DATE'] + other_cols)
    df_copy = pd.concat([df_base, df_copy])
    df_copy[df_copy.columns[1:]] = df_copy[df_copy.columns[1:]].fillna(0).astype(int)

    df_copy['weekend']=[1 if i>=5 else 0 for i in df_copy['DATE'].dt.weekday]
    df_copy['is_spring'] = df_copy['DATE'].apply(is_spring).astype(int)
    df_copy['is_summer'] = df_copy['DATE'].apply(is_summer).astype(int)
    df_copy['is_autumn'] = df_copy['DATE'].apply(is_autumn).astype(int)
    df_copy['is_winter'] = df_copy['DATE'].apply(is_winter).astype(int)
    #df_copy['date']=df_copy['Tarih'].dt.date
    #df_copy['date']=pd.to_datetime(df_copy['date'])
    return df_copy


from lazyauto.preperation.drop_outlier import drop_outliers
from lazyauto.Eda.eda import outlier_dedection
from sklearn import datasets
from lightgbm import LGBMRegressor
import pandas as pd

diab=datasets.load_diabetes(as_frame=True)
df=pd.DataFrame(diab['frame'],columns=diab['feature_names']+['target'])


def test_drop_outlier():
    outlier_index=outlier_dedection(df, target='target', model=LGBMRegressor(),contaminations=[0.1])
    before_drop=df.shape[0]
    dropped_df=drop_outliers(df,outlier_index).shape[0]
    assert before_drop>dropped_df
    